/*
    mtr  --  a network diagnostic tool
    Copyright (C) 1997,1998  Matt Kimball

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
    
   1999-08-13 ok Olav@okvittem.priv.no  added -psize

*/

#include <config.h>

#if defined(HAVE_SYS_XTI_H)
#include <sys/xti.h>
#endif

#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#ifdef ENABLE_IPV6
#include <netinet/ip6.h>
#include <netinet/icmp6.h>
#endif
#include <memory.h>
#include <unistd.h>
#include <stdio.h>
#include <math.h>
#include <errno.h>
#include <netdb.h>

#include "net.h"
#include "display.h"

#ifndef ENABLE_IPV6
#define sockaddr_storage sockaddr
#undef ss_len
#define ss_len sa_len
#undef ss_family
#define ss_family sa_family
#endif

#define ICMP6ECHOLEN	8


#define MaxTransit 4

/*  We can't rely on header files to provide this information, because
    the fields have different names between, for instance, Linux and 
    Solaris  */
struct ICMPHeader {
  uint8 type;
  uint8 code;
  uint16 checksum;
  uint16 id;
  uint16 sequence;
};

/*  Structure of an IP header.  */
struct IPHeader {
  uint8 version;
  uint8 tos;
  uint16 len;
  uint16 id;
  uint16 frag;
  uint8 ttl;
  uint8 protocol;
  uint16 check;
  uint32 saddr;
  uint32 daddr;
};
  
#define ICMP_ECHO		8
#define ICMP_ECHOREPLY		0

#define ICMP_TSTAMP		13
#define ICMP_TSTAMPREPLY	14

#define ICMP_TIME_EXCEEDED	11

#ifndef SOL_IP
#define SOL_IP 0
#endif

#define saddr_correction(addr) BSDfix ? addr : 0

struct nethost {
  union {
    struct sockaddr_in _addr4;
#define addr4 _addr._addr4.sin_addr
#ifdef ENABLE_IPV6
    struct sockaddr_in6 _addr6;
#define addr6 _addr._addr6.sin6_addr
#endif
    struct sockaddr _addr0;
#define addr0 _addr._addr0
  } _addr;
  int xmit;
  int returned;
  int sent;
  int up;
  long long total;
  int last;
  int best;
  int worst;
  int transit;
  int saved[SAVED_PINGS];
  int saved_seq_offset;
};

struct sequence {
    int index;
    int transit;
    int saved_seq;
    struct timeval time;
};


/* Configuration parameter: How many queries to unknown hosts do we
   send? (This limits the amount of traffic generated if a host is not
   reachable) */
#define MAX_UNKNOWN_HOSTS 5


static struct nethost host[MaxHost];
static struct sequence sequence[MaxSequence];
static struct timeval reset = { 0, 0 };

int sendsock4 = -1, recvsock4 = -1;
#ifdef ENABLE_IPV6
int sendsock6 = -1, recvsock6 = -1;
#endif
static struct sockaddr_storage sourceaddress;
static struct sockaddr_storage remoteaddress;
static int batch_at = 0;


extern int packetsize;
static int numhosts = 10;

/* return the number of microseconds to wait before sending the next
   ping */
int calc_deltatime (float waittime)
{
  waittime /= numhosts;
  return 1000000 * waittime;
}


/* This doesn't work for odd sz. I don't know enough about this to say
   that this is wrong. It doesn't seem to cripple mtr though. -- REW */
int checksum(void *data, int sz) {
  unsigned short *ch;
  unsigned int sum;

  sum = 0;
  ch = data;
  sz = sz / 2;
  while(sz--) {
    sum += *(ch++);
  }
  
  sum = (sum >> 16) + (sum & 0xffff);  

  return (~sum & 0xffff);  
}


static int BSDfix = 0;

int new_sequence(int index) {
  static int next_sequence = 0;
  int seq;

  seq = next_sequence++;
  if(next_sequence >= MaxSequence)
    next_sequence = 0;

  sequence[seq].index = index;
  sequence[seq].transit = 1;
  sequence[seq].saved_seq = ++host[index].xmit;
  memset(&sequence[seq].time, 0, sizeof(sequence[seq].time));
  
  host[index].transit = 1;
  if (host[index].sent)
    host[index].up = 0;
  host[index].sent = 1;
  net_save_xmit(index);
  
  return seq;
}

/*  Attempt to find the host at a particular number of hops away  */
void net4_send_query(int index) {
  /*ok  char packet[sizeof(struct IPHeader) + sizeof(struct ICMPHeader)];*/
  char packet[MAXPACKET];
  struct IPHeader *ip;
  struct ICMPHeader *icmp;

  /*ok  int packetsize = sizeof(struct IPHeader) + sizeof(struct ICMPHeader) + datasize;*/
  int rv;
  static int first=1;

  if ( sendsock4 < 0 )
    return;

  if ( packetsize < MINPACKET ) packetsize = MINPACKET;
  if ( packetsize > MAXPACKET ) packetsize = MAXPACKET;
  memset(packet, 0, packetsize);

  ip = (struct IPHeader *)packet;
  icmp = (struct ICMPHeader *)(packet + sizeof(struct IPHeader));

  ip->version = 0x45;
  ip->tos = 0;
  ip->len = BSDfix ? packetsize: htons (packetsize);
  ip->id = 0;
  ip->frag = 0;
  ip->ttl = index + 1;
  ip->protocol = IPPROTO_ICMP;
  /* BSD needs the source address here, Linux & others do not... */
  ip->saddr = saddr_correction(((struct sockaddr_in *)&sourceaddress)->sin_addr.s_addr);
  ip->daddr = ((struct sockaddr_in *)&remoteaddress)->sin_addr.s_addr;

  icmp->type = ICMP_ECHO;
  icmp->id = getpid();
  icmp->sequence = new_sequence(index);

  icmp->checksum = checksum(icmp, packetsize - sizeof(struct IPHeader));
  ip->check = checksum(ip, packetsize);

  gettimeofday(&sequence[icmp->sequence].time, NULL);
  rv = sendto(sendsock4, packet, packetsize, 0, 
	 (struct sockaddr *)&remoteaddress,
#ifdef linux
	 SA_LEN((struct sockaddr *)&remoteaddress)
#else
	 remoteaddress.ss_len
#endif
	 );
  if (first && (rv < 0) && (errno == EINVAL)) {
    ip->len = packetsize;
    rv = sendto(sendsock4, packet, packetsize, 0, 
		(struct sockaddr *)&remoteaddress,
#ifdef linux
		SA_LEN((struct sockaddr *)&remoteaddress)
#else
		remoteaddress.ss_len
#endif
		);
    if (rv >= 0) {
      fprintf (stderr, "You've got a broken (FreeBSD?) system\n");
      BSDfix = 1;
    }
  }
  if (rv < 0) {
    perror("sendto");
    exit(1);
  }
  first = 0;
}

#ifdef ENABLE_IPV6
void net6_send_query(int index) {
  /*ok  char packet[sizeof(struct IPHeader) + sizeof(struct ICMPHeader)];*/
  char packet[MAXPACKET];
  struct icmp6_hdr *icmp6;

  /*ok  int packetsize = sizeof(struct IPHeader) + sizeof(struct ICMPHeader) + datasize;*/
  int rv;
  static int first=1;
  int hops;

  if (sendsock6 < 0)
    return;

  if (packetsize < MINPACKET) packetsize = MINPACKET;
  if (packetsize > MAXPACKET) packetsize = MAXPACKET;
  memset(packet, 0, packetsize);

  hops = index + 1;
  if (setsockopt(sendsock6, IPPROTO_IPV6, IPV6_UNICAST_HOPS,
	&hops, sizeof(hops)) < 0) {
    perror("setsockopt(IPV6_UNICAST_HOPS)");
    exit(1);
  }

  icmp6 = (struct icmp6_hdr *)packet;

  icmp6->icmp6_type = ICMP6_ECHO_REQUEST;
  icmp6->icmp6_code = 0;
  icmp6->icmp6_cksum = 0;
  icmp6->icmp6_seq = new_sequence(index);
  icmp6->icmp6_id = getpid() & 0xffff;

  gettimeofday(&sequence[icmp6->icmp6_seq].time, NULL);
  rv = sendto(sendsock6, packet, packetsize, 0, 
	 (struct sockaddr *)&remoteaddress,
#ifdef linux
	 SA_LEN((struct sockaddr *)&remoteaddress)
#else
	 remoteaddress.ss_len
#endif
	);
  if (rv < 0) {
    perror("sendto");
    exit(1);
  }
  first = 0;
}
#endif

void net_send_query(int index) {
  switch (remoteaddress.ss_family) {
  case AF_INET:
    net4_send_query(index);
    break;
#ifdef ENABLE_IPV6
  case AF_INET6:
    net6_send_query(index);
    break;
#endif
  }
}

/*   We got a return on something we sent out.  Record the address and
     time.  */
void net4_process_ping(int seq, uint32 addr, struct timeval now) {
  int index;
  int totusec;

  if(sendsock4 < 0)
    return;

  if(seq < 0 || seq >= MaxSequence)
    return;

  if(!sequence[seq].transit)
    return;
  sequence[seq].transit = 0;

  index = sequence[seq].index;

  totusec = (now.tv_sec  - sequence[seq].time.tv_sec ) * 1000000 +
            (now.tv_usec - sequence[seq].time.tv_usec);

  if(host[index].addr4.s_addr == 0) {
    host[index].addr0.sa_family = AF_INET;
#ifndef linux
    host[index].addr0.sa_len = sizeof(struct sockaddr_in);
#endif
    host[index].addr4.s_addr = addr;
    display_rawhost(index, &host[index].addr0);
  }
  if(host[index].returned <= 0) {
    host[index].best = host[index].worst = totusec;
  }
  host[index].last = totusec;
  if(totusec < host[index].best)
    host[index].best = totusec;
  if(totusec > host[index].worst)
    host[index].worst = totusec;

  host[index].total += totusec;
  host[index].returned++;
  host[index].sent = 0;
  host[index].up = 1;
  host[index].transit = 0;

  net_save_return(index, sequence[seq].saved_seq, totusec);
  display_rawping(index, totusec);
}

#ifdef ENABLE_IPV6
void net6_process_ping(int seq, struct in6_addr *addr, struct timeval now) {
  int index;
  int totusec;

  if(sendsock6 < 0)
    return;

  if(seq < 0 || seq >= MaxSequence)
    return;

  if(!sequence[seq].transit)
    return;
  sequence[seq].transit = 0;

  index = sequence[seq].index;

  totusec = (now.tv_sec  - sequence[seq].time.tv_sec ) * 1000000 +
            (now.tv_usec - sequence[seq].time.tv_usec);

  if(IN6_IS_ADDR_UNSPECIFIED(&host[index].addr6)) {
    host[index].addr0.sa_family = AF_INET6;
#ifndef linux
    host[index].addr0.sa_len = sizeof(struct sockaddr_in6);
#endif
    host[index].addr6 = *addr;
    display_rawhost(index, &host[index].addr0);
  }
  if(host[index].returned <= 0) {
    host[index].best = host[index].worst = totusec;
  }
  host[index].last = totusec;
  if(totusec < host[index].best)
    host[index].best = totusec;
  if(totusec > host[index].worst)
    host[index].worst = totusec;

  host[index].total += totusec;
  host[index].returned++;
  host[index].sent = 0;
  host[index].up = 1;
  host[index].transit = 0;

  net_save_return(index, sequence[seq].saved_seq, totusec);
  display_rawping(index, totusec);
}
#endif

/*  We know a packet has come in, because the main select loop has called us,
    now we just need to read it, see if it is for us, and if it is a reply 
    to something we sent, then call net4_process_ping()  */
void net4_process_return() {
  char packet[2048];
  struct sockaddr_in fromaddr;
  int fromaddrsize;
  int num;
  struct ICMPHeader *header;
  struct timeval now;

  if(recvsock4 < 0) return;

  gettimeofday(&now, NULL);

  fromaddrsize = sizeof(fromaddr);
  num = recvfrom(recvsock4, packet, 2048, 0, 
		 (struct sockaddr *)&fromaddr, &fromaddrsize);

  if(num < sizeof(struct IPHeader) + sizeof(struct ICMPHeader))
    return;

  header = (struct ICMPHeader *)(packet + sizeof(struct IPHeader));
  if(header->type == ICMP_ECHOREPLY) {
    if(header->id != (uint16)getpid())
      return;

    net4_process_ping(header->sequence, fromaddr.sin_addr.s_addr, now);
  } else if(header->type == ICMP_TIME_EXCEEDED) {
    if(num < sizeof(struct IPHeader) + sizeof(struct ICMPHeader) + 
             sizeof(struct IPHeader) + sizeof(struct ICMPHeader))
      return;
    
    header = (struct ICMPHeader *)(packet + sizeof(struct IPHeader) + 
				sizeof(struct ICMPHeader) + sizeof(struct IPHeader));
    if(header->id != (uint16)getpid())
      return;

    net4_process_ping(header->sequence, fromaddr.sin_addr.s_addr, now);
  }
}

#ifdef ENABLE_IPV6
void net6_process_return() {
  char packet[2048];
  struct sockaddr_storage fromaddr;
  int fromaddrsize;
  int num;
  struct icmp6_hdr *header;
  struct timeval now;

  if(recvsock6 < 0) return;

  gettimeofday(&now, NULL);

  fromaddrsize = sizeof(fromaddr);
  num = recvfrom(recvsock6, packet, 2048, 0, 
		 (struct sockaddr *)&fromaddr, &fromaddrsize);

  if (num < sizeof(struct icmp6_hdr))
    return;

  header = (struct icmp6_hdr *)packet;
  if(header->icmp6_type == ICMP6_ECHO_REPLY) {
    if(header->icmp6_id != (uint16)(getpid() & 0xffff))
      return;

    net6_process_ping(header->icmp6_seq,
      &((struct sockaddr_in6 *)&fromaddr)->sin6_addr, now);
  } else if (header->icmp6_type == ICMP6_TIME_EXCEEDED) {
    if(num < sizeof(struct icmp6_hdr) + 
             sizeof(struct ip6_hdr) + sizeof(struct icmp6_hdr))
      return;
    
    header = (struct icmp6_hdr *)(packet + sizeof(struct icmp6_hdr)
					    + sizeof(struct ip6_hdr));
    if(header->icmp6_id != (getpid() & 0xffff))
      return;

    net6_process_ping(header->icmp6_seq,
      &((struct sockaddr_in6 *)&fromaddr)->sin6_addr, now);
  }
}
#endif

void net_process_return() {
  switch (remoteaddress.ss_family) {
  case AF_INET:
    net4_process_return();
    break;
#ifdef ENABLE_IPV6
  case AF_INET6:
    net6_process_return();
    break;
#endif
  }
}

const struct sockaddr *net_addr(int at) {
  if (unspecaddr(at))
    return NULL;
  else
    return &host[at].addr0;
}

int net_percent(int at) {
  if((host[at].xmit - host[at].transit) == 0) 
    return 0;

  return 100 - (100 * host[at].returned / (host[at].xmit - host[at].transit));
}

int net_last(int at) {
  return host[at].last;
}

int net_best(int at) {
  return host[at].best;
}

int net_worst(int at) {
  return host[at].worst;
}

int net_avg(int at) {
  if(host[at].returned == 0)
    return 0;

  return host[at].total / host[at].returned;
}

static int finaladdr(int at) {
  switch (remoteaddress.ss_family) {
  case AF_INET:
    if (host[at].addr4.s_addr == ((struct sockaddr_in *)&remoteaddress)->sin_addr.s_addr)
      return 1;
    break;
#ifdef ENABLE_IPV6
  case AF_INET6:
    if (IN6_ARE_ADDR_EQUAL(&host[at].addr6, &((struct sockaddr_in6 *)&remoteaddress)->sin6_addr))
      return 1;
    break;
#endif
  }
  return 0;
}

static int unspecaddr(int at) {
  switch (remoteaddress.ss_family) {
  case AF_INET:
    if (host[at].addr4.s_addr == 0)
      return 1;
    break;
#ifdef ENABLE_IPV6
  case AF_INET6:
    if (IN6_IS_ADDR_UNSPECIFIED(&host[at].addr6))
      return 1;
    break;
#endif
  }
  return 0;
}

int net_max() {
  int at;
  int max;

  max = 0;
  for(at = 0; at < MaxHost-2; at++) {
    if(finaladdr(at)) {
      return at + 1;
    } else if(!unspecaddr(at)) {
      max = at + 2;
    }
  }

  return max;
}


/* Added by Brian Casey December 1997 bcasey@imagiware.com*/
int net_returned(int at) { 
   return host[at].returned;
}
int net_xmit(int at) { 
   return host[at].xmit;
}
int net_transit(int at) { 
   return host[at].transit;
}

int net_up(int at) { 
   return host[at].up;
}

void net_end_transit() {
  int at;

  for(at = 0; at < MaxHost; at++) {
    host[at].transit = 0;
  }
}



int net_send_batch() {
  int n_unknown, i;

  net_send_query(batch_at);

  n_unknown = 0;

  for (i=0;i<batch_at;i++) {
    if (unspecaddr(i))
      n_unknown++;
    if (finaladdr(i))
      n_unknown = 100; /* Make sure we drop into "we should restart" */
  }

  if (finaladdr(batch_at) ||
      (n_unknown > MAX_UNKNOWN_HOSTS) ||
      (batch_at >= MaxHost-2)) {
    numhosts = batch_at+1;
    batch_at = 0;
    return 1;
  }

  batch_at++;
  return 0;
}


int net_preopen() {
  int trueopt = 1;

  sendsock4 = socket(AF_INET, SOCK_RAW, IPPROTO_RAW);

#ifdef IP_HDRINCL
  /*  FreeBSD wants this to avoid sending out packets with protocol type RAW
      to the network.  */
  if(setsockopt(sendsock4, SOL_IP, IP_HDRINCL, &trueopt, sizeof(trueopt)))
  {
    perror("setsockopt(IP_HDRINCL,1)");
    sendsock4 = -1;
  }
#endif

  if(sendsock4 >= 0)
    recvsock4 = socket(AF_INET, SOCK_RAW, IPPROTO_ICMP);
  if(recvsock4 < 0) {
    close(sendsock4);
    sendsock4 = -1;
  }

#ifdef ENABLE_IPV6
  sendsock6 = socket(AF_INET6, SOCK_RAW, IPPROTO_ICMPV6);

  if(sendsock6 >= 0)
    recvsock6 = socket(AF_INET6, SOCK_RAW, IPPROTO_ICMPV6);
  if(recvsock6 < 0) {
    close(sendsock6);
    sendsock6 = -1;
  }

  /* Make sure we have either a valid IPV4 or IPV6 socket pair. */
  if(((recvsock4 < 0) || (sendsock4 < 0)) &&
     ((recvsock6 < 0) || (sendsock6 < 0)))
    return -1;
#else
  /* Make sure we have either a valid IPV4 socket pair. */
  if((recvsock4 < 0) || (sendsock4 < 0))
    return -1;
#endif

  return 0;
}
 
int net4_open(int addr) {
  struct sockaddr_in *sin;

  if(sendsock4 < 0)
    return(-1);

  net_reset();
#if 0
  if (sendsock6 >= 0)
    close(sendsock6);
  sendsock6 = -1;
  if (recvsock6 >= 0)
    close(recvsock6);
  recvsock6 = -1;
#endif

  memset(&remoteaddress, 0, sizeof(remoteaddress));
  sin = (struct sockaddr_in *)&remoteaddress;
  sin->sin_family = AF_INET;
#ifndef linux
  sin->sin_len = sizeof(struct sockaddr_in);
#endif
  sin->sin_addr.s_addr = addr;

  return 0;
}

void net4_reopen(int addr) {
  struct sockaddr_in *sin;
  int at;

  for(at = 0; at < MaxHost; at++) {
    memset(&host[at], 0, sizeof(host[at]));
  }

  memset(&remoteaddress, 0, sizeof(remoteaddress));
  sin = (struct sockaddr_in *)&remoteaddress;
  sin->sin_family = AF_INET;
#ifndef linux
  sin->sin_len = sizeof(struct sockaddr_in);
#endif
  sin->sin_addr.s_addr = addr;

  net_reset ();
  net_send_batch();
}

#ifdef ENABLE_IPV6
int net6_open(struct in6_addr *addr) {
  struct sockaddr_in6 *sin6;

  if(sendsock6 < 0)
    return(-1);

  net_reset();
#if 0
  if (sendsock4 >= 0)
    close(sendsock4);
  sendsock4 = -1;
  if (recvsock4 >= 0)
    close(recvsock4);
  recvsock4 = -1;
#endif

  memset(&remoteaddress, 0, sizeof(remoteaddress));
  sin6 = (struct sockaddr_in6 *)&remoteaddress;
  sin6->sin6_family = AF_INET6;
#ifndef linux
  sin6->sin6_len = sizeof(struct sockaddr_in6);
#endif
  memcpy(&sin6->sin6_addr, addr, sizeof(*addr));

  return 0;
}

void net6_reopen(struct in6_addr *addr) {
  struct sockaddr_in6 *sin6;
  int at;

  if(sendsock6 < 0) return;

  for(at = 0; at < MaxHost; at++) {
    memset(&host[at], 0, sizeof(host[at]));
  }

  memset(&remoteaddress, 0, sizeof(remoteaddress));
  sin6 = (struct sockaddr_in6 *)&remoteaddress;
  sin6->sin6_family = AF_INET6;
#ifndef linux
  sin6->sin6_len = sizeof(struct sockaddr_in6);
#endif
  memcpy(&sin6->sin6_addr, addr, sizeof(*addr));

  net_reset ();
  net_send_batch();
}
#endif

void net_reset() {
  int at;
  int i;

  batch_at = 0;
  numhosts = 10;

  for(at = 0; at < MaxHost; at++) {
    host[at].xmit = 0;
    host[at].transit = 0;
    host[at].returned = 0;
    host[at].sent = 0;
    host[at].up = 0;
    host[at].total = 0;
    host[at].best = 0;
    host[at].worst = 0;
    for (i=0; i<SAVED_PINGS; i++) {
      host[at].saved[i] = -2;	/* unsent */
    }
    host[at].saved_seq_offset = -SAVED_PINGS+2;
  }
  
  for(at = 0; at < MaxSequence; at++) {
    sequence[at].transit = 0;
  }

  gettimeofday(&reset, NULL);
}

void net_close() {
  if (sendsock4 >= 0)
    close(sendsock4);
  if (recvsock4 >= 0)
    close(recvsock4);
#ifdef ENABLE_IPV6
  if (sendsock6 >= 0)
    close(sendsock6);
  if (recvsock6 >= 0)
    close(recvsock6);
#endif
}

int net4_waitfd() {
  return recvsock4;
}

#ifdef ENABLE_IPV6
int net6_waitfd() {
  return recvsock6;
}
#endif

int net_waitfd() {
  switch (remoteaddress.ss_family) {
  case AF_INET:
    return net4_waitfd();
#ifdef ENABLE_IPV6
  case AF_INET6:
    return net6_waitfd();
#endif
  default:
    return -1;
  }
}


int* net_saved_pings(int at) {
	return host[at].saved;
}

void net_save_increment() 
{
  int at;
  for (at = 0; at < MaxHost; at++) {
    memmove(host[at].saved, host[at].saved+1, (SAVED_PINGS-1)*sizeof(int));
    host[at].saved[SAVED_PINGS-1] = -2;
    host[at].saved_seq_offset += 1;
  }
}

void net_save_xmit(int at) {
  if (host[at].saved[SAVED_PINGS-1] != -2) 
    net_save_increment();
  host[at].saved[SAVED_PINGS-1] = -1;
}

void net_save_return(int at, int seq, int ms) {
	int idx;
	idx = seq - host[at].saved_seq_offset;
	if (idx < 0 || idx > SAVED_PINGS) {
	  return;
	}
	host[at].saved[idx] = ms;
}
